/* 
 *  Driver for Frontend Transport Stream playout
 *
 *  Written by Gert-Jan van Dijk <aka_mr_zapper@hotmail.com>
 *           & Emile Meeuse <crotcheteer@hotmail.com>
 *             Ordina Technical Automation
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.=
 */    


#include <linux/module.h>   /* Required header, must be included by a module source */
#include <linux/kernel.h>   /* We're in kernel space */
#include <linux/sched.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/vmalloc.h>
#include <linux/slab.h>
#include <asm/uaccess.h>

#include <include/linux/dvb/frontend.h>
#include "dvb_fets.h"

#ifdef _DVB_NEW_ // Definitions and paths for DVB driver 1.1.0 and higher
    #define DMXDEV struct dmxdev
    #define DMXDEV_BUFFER struct dmxdev_buffer
    #define DMX_FRONTEND struct dmx_frontend
    #include <../build-2.4/dvb_net.h>
    #include <../build-2.4/dmxdev.h>
    #include <../build-2.4/dvb_demux.h>
#else            // Definitions and paths for DVB driver 1.0.1 and lower
    #define DMXDEV dmxdev_t
    #define DMXDEV_BUFFER dmxdev_buffer_t
    #define DMX_FRONTEND dmx_frontend_t
    #include <driver/dvb_demux.h>
    #include <driver/dmxdev.h>
    #include <driver/dvb_net.h>
#endif

extern struct dvb_device* dvbdev_find_device (int minor);
                                
/*
 *  Prototypes
 */

static int  dvb_fets_open (struct inode *inode, struct file *file);
static int  dvb_fets_release (struct inode *inode, struct file *file);
static int  dvb_fets_ioctl(struct inode* inode, struct file* filp, unsigned int cmd, unsigned long arg);
static int  dvb_fets_write(struct file *filp, const char *buf, size_t count, loff_t *f_pos);

/*
 *  Global data 
 */

/* Whenever an operation is performed on a character device file
   associated with the major number (as received from register_chrdev()), the 
   kernel finds and invokes the proper function from this table. */
static struct file_operations dvb_fets_fops =
{
    .owner      = THIS_MODULE,
    .ioctl      = dvb_fets_ioctl,
    .write      = dvb_fets_write,
    .open       = dvb_fets_open,
    .release    = dvb_fets_release,
};

static struct dvb_frontend_info dvb_fets_info = 
{
    .name                   = "DVB-T TS frontend",
    .type                   = (fe_type_t)DVB_TS,
    .frequency_min          = 0,
    .frequency_max          = 863250000,
    .frequency_stepsize     = 62500,
    /*.frequency_tolerance         = */    /* FIXME: 12% of SR */
    .symbol_rate_min        = 0,           /* FIXME */
    .symbol_rate_max        = 9360000,     /* FIXME */
    .symbol_rate_tolerance  = 4000,
    .notifier_delay         = 0,
    .caps                   = FE_CAN_FEC_1_2 | FE_CAN_FEC_2_3 | FE_CAN_FEC_3_4 |
                              FE_CAN_FEC_4_5 | FE_CAN_FEC_5_6 | FE_CAN_FEC_6_7 |
                              FE_CAN_FEC_7_8 | FE_CAN_FEC_8_9 | FE_CAN_FEC_AUTO |
                              FE_CAN_QAM_16 | FE_CAN_QAM_64 | FE_CAN_QAM_AUTO |
                              FE_CAN_TRANSMISSION_MODE_AUTO |
                              FE_CAN_GUARD_INTERVAL_AUTO |
                              FE_CAN_HIERARCHY_AUTO,
};

/*
 *  Local data 
 */

static int      dvb_fets_major      = 0; /* = 0, use dynamic assignment */
static char*    dvb_fets_name       = "dvb_fets";
static int      dvb_frontend_debug  = 1;
#define dprintk if (dvb_frontend_debug) printk

/*
 *  Type definitions
 */

#define MAX_EVENT 8

struct fets_events
{
	struct dvb_frontend_event events[MAX_EVENT];
	int                       eventw;
	int                       eventr;
	int                       overflow;
	wait_queue_head_t         wait_queue;
	struct semaphore          sem;
};

struct fets_info
{
    /* devices */
    struct dvb_device       dvb_dev;
    struct dvb_net          dvb_net;

    struct saa7146_dev	    *dev;

    struct dvb_i2c_bus	    *i2c_bus;	
    struct budget_info	    *card;

    unsigned char		    *grabbing;
//    struct saa7146_pgtable  pt;

    struct tasklet_struct   fidb_tasklet;
    struct tasklet_struct   vpe_tasklet;

    DMXDEV                  dmxdev;
    DMXDEV                  *pdmxdev;
    
    struct dvb_demux	    demux;

    DMX_FRONTEND            hw_frontend;
    DMX_FRONTEND            mem_frontend;

    int                     fe_synced; 
    struct semaphore        pid_mutex;

    u8                      tsf;
    u32                     ttbp;
    int                     feeding;

    struct dvb_adapter      *dvb_adapter;
    void                    *priv;

    unsigned long           symb_rate; /* kilo symbols per sec */
    unsigned char*          buf;
    struct semaphore        sem;
	struct fets_events      events;
};

/*
 *  Initialization function:
 *  The kernel calls "module_init" to initialize a newly loaded module
 *  (with insmod, which also does the dynamic linking).
 *      - Get the major number
 *      - 
 *
 */
static int __init dvb_fets_init (void)
{
    int result;

    printk("%s\n",__FUNCTION__);

    result = register_chrdev(dvb_fets_major, dvb_fets_name, &dvb_fets_fops);
    if (result < 0) 
    {
        printk(KERN_WARNING "dvb_fets: can't register major %d\n",dvb_fets_major);
        return result;
    }
    if (dvb_fets_major == 0) dvb_fets_major = result; /* dynamic */

    return 0;
}

/*
 *  Cleanup function:
 *  The kernel calls "module_exit" just before module removal (unloading).
 *      - Release the major number
 */
static void __exit dvb_fets_cleanup (void)
{
    printk("%s\n",__FUNCTION__);

    if (unregister_chrdev(dvb_fets_major, dvb_fets_name) < 0)
    {
        printk(KERN_WARNING "dvb_fets: can't unregister major %d\n",dvb_fets_major);
    }
}

static DMX_FRONTEND *get_fe(struct dmx_demux *demux, int type)
{
	struct list_head *head, *pos;

	head=demux->get_frontends(demux);
	if (!head)
		return 0;
	list_for_each(pos, head)
		if (DMX_FE_ENTRY(pos)->source==type)
			return DMX_FE_ENTRY(pos);
	
	return 0;
}

static inline void dvb_dmxdev_buffer_init(DMXDEV_BUFFER *buffer) 
{
	buffer->data=0;
	buffer->size=8192;
	buffer->pread=0;
	buffer->pwrite=0;
	buffer->error=0;
	init_waitqueue_head(&buffer->queue);
}

/*
 *  Open function:
 *      - Increment the usage count
 *      - Check for device specific errors
 *      - Initialize the device, if opened for first time
 *      - Identify minor number and update f_op pointer if necessary
 *      - Allocate and fill any data struct to be put in filp->private_data
 */
static
int dvb_fets_open (struct inode *inode, struct file *file)
{
    struct dvb_device   *dvb;
    DMX_FRONTEND        *front;
    struct fets_info    *fets;
    int                 result = 0;

    dprintk("%s\n",__FUNCTION__);

    MOD_INC_USE_COUNT;

    file->private_data = kmalloc(sizeof(struct fets_info), GFP_KERNEL);
    if (!file->private_data)
    {
        printk(KERN_ERR "dvb_fets: kalloc failed\n");        
        result = -ENOMEM;
    }
    memset(file->private_data,0,sizeof(struct fets_info));
    fets = (struct fets_info *) file->private_data;
    sema_init(&fets->sem, 1);
    dvb = dvbdev_find_device(DVB_DEVICE_DEMUX);
    fets->pdmxdev = dvb->priv;

    if (down_interruptible (&fets->sem))
		return -ERESTARTSYS;
	
    if ((file->f_flags&O_ACCMODE)==O_RDWR) { 
		fets->pdmxdev->dvr_orig_fe=fets->pdmxdev->demux->frontend;
		
		if (!fets->pdmxdev->demux->write) {
			up(&fets->sem);
			return -EOPNOTSUPP;
		}
		
		front=get_fe(fets->pdmxdev->demux, DMX_MEMORY_FE);
	
		if (!front) {
			up(&fets->sem);
			return -EINVAL;
		}
		fets->pdmxdev->demux->disconnect_frontend(fets->pdmxdev->demux);	
		fets->pdmxdev->demux->connect_frontend(fets->pdmxdev->demux, front);	
	}
	up(&fets->sem);
	return 0;
}

/*
 *  Release function:
 *      - Deallocate anything that open() allocated in filp->private_data
 *      - Shut down the device 
 *      - Decrement usage count
 */
static int dvb_fets_release(struct inode *inode, struct file *file)
{
    struct fets_info *fets;

    dprintk("%s\n",__FUNCTION__);

    fets = (struct fets_info *) file->private_data;
    kfree(file->private_data);

    MOD_DEC_USE_COUNT;
    
    if (down_interruptible (&fets->sem))
		return -ERESTARTSYS;

	if ((file->f_flags&O_ACCMODE)==O_WRONLY) {
		fets->pdmxdev->demux->disconnect_frontend(fets->pdmxdev->demux);	
		fets->pdmxdev->demux->connect_frontend(fets->pdmxdev->demux, 
						fets->pdmxdev->dvr_orig_fe);
	}
	if ((file->f_flags&O_ACCMODE)==O_RDONLY) {
		if (fets->pdmxdev->dvr_buffer.data) {
			void *mem=fets->pdmxdev->dvr_buffer.data;
			mb();
			spin_lock_irq(&fets->pdmxdev->lock);
			fets->pdmxdev->dvr_buffer.data=0;
			spin_unlock_irq(&fets->pdmxdev->lock);
			vfree(mem);
		}
	}
	up(&fets->sem);
    return 0;
}


/*
 *  Write:
 */
static int dvb_fets_write(struct file *file, const char *buf, size_t count, loff_t *f_pos)
{
    struct fets_info *fets = (struct fets_info *) file->private_data;

    int ret = count;

    if (down_interruptible (&fets->sem))
        return -ERESTARTSYS;

    if (!fets->pdmxdev->demux->write)
		return -EOPNOTSUPP;

    /* Allocate memory */
    fets->buf = kmalloc(count * sizeof(unsigned char), GFP_KERNEL);
    if (!fets->buf)
    {
        dprintk("ENOMEM\n");
        ret = -ENOMEM;
        goto out1;
    }

    /* Copy from user space to kernel space */
    if (copy_from_user(fets->buf, buf, count))
    {
        dprintk("EFAULT\n");        
        ret = -EFAULT;
        goto out2;
    }
    
    ret = fets->pdmxdev->demux->write(fets->pdmxdev->demux, fets->buf, count);
    
    dprintk("%s: Stored %d bytes\n",__FUNCTION__,ret);    

    if (signal_pending(current))
        ret = -EINTR;
out2:
    kfree(fets->buf);
out1: 
    up(&fets->sem);
    return ret;
}


/*
 *  IO Control function:
 *      -
 */
static int dvb_fets_ioctl(struct inode* inode, struct file* filp, unsigned int cmd, unsigned long arg)
{
    dprintk("%s\n",__FUNCTION__);

    switch (cmd) 
    {
    case FE_SET_SR:
    {
        struct fets_info* fets              = filp->private_data;
        struct dvb_frontend_ts* ts_fets_p   = (struct dvb_frontend_ts*) arg;

        fets->symb_rate   = ts_fets_p->symb_rate;

        break;
    }

    case FE_GET_INFO:
    {
        memcpy ( (void*)arg,
                 &dvb_fets_info,
                 sizeof(struct dvb_frontend_info));
    
        break;
    }

    case FE_READ_STATUS:
    {
        fe_status_t *status = (fe_status_t*)arg;
        *status = FE_HAS_SIGNAL
                | FE_HAS_CARRIER
                | FE_HAS_VITERBI
                | FE_HAS_SYNC
                | FE_HAS_LOCK;
        break;
    }

    case FE_READ_BER:
    {
        u32 *ber = (u32 *) arg;
        *ber = 0;
        break;
    }

    case FE_READ_SIGNAL_STRENGTH:
    {
        u8 signal = 0xff;
        *((u16*) arg) = (signal << 8) | signal;
        break;
    }

    case FE_READ_SNR:
    {
        u8 snr = 0xf0;
        *(u16*) arg = (snr << 8) | snr;
        break;
    }

    case FE_READ_UNCORRECTED_BLOCKS: 
        *(u32*) arg = 0;
        break;

    case FE_SET_FRONTEND:
    case FE_GET_FRONTEND:
    case FE_SET_TONE:
    case FE_SET_VOLTAGE:
        break;

    case FE_GET_EVENT:
    {
        struct dvb_frontend_event * event = (struct dvb_frontend_event *)arg;
        event->status = FE_HAS_SIGNAL
                      | FE_HAS_CARRIER
                      | FE_HAS_VITERBI
                      | FE_HAS_SYNC
                      | FE_HAS_LOCK;
        break;
    }

    default:
        return -EOPNOTSUPP;
    }
    return 0;
}

/*
 *  Mark modules initialization and cleanup functions 
 */
module_init(dvb_fets_init);
module_exit(dvb_fets_cleanup);

/* 
 *  Place documentation on the module in the object file 
 */
MODULE_DESCRIPTION("DVB TS Frontend");
MODULE_AUTHOR("Gert-Jan van Dijk / Emile Meeuse");
MODULE_LICENSE("GPL");


